--[[
	Template encoding code is heavily borrowed from wowhead, but use a different map.
	StringToTemplate and TemplateToString
]]

local stop = 'Z'
local talented_map = "012345abcdefABCDEFmnopqrMNOPQRtuvwxy"
local classmap = { 'DRUID','HUNTER','MAGE','PALADIN','PRIEST','ROGUE','SHAMAN','WARLOCK','WARRIOR' }

function Talented:StringToTemplate(code, class, nmap, template)
	nmap = nmap or talented_map
	if code:len() <= 0 then return end
	
	local index = math.modf((nmap:find(code:sub(1, 1), nil, true) - 1) / 3) + 1
	assert(index and index <= #classmap, "Unknown class code")
	
--[[
	if not class then
		_, class = UnitClass("player")
	end
	assert(classmap[index] == class, "Invalid class")
]]
	local class = classmap[index]
	
	template = template or {}
	
	local next = function()
		local t = {}
		template[#template + 1] = t
		return t
	end
	
	local add = function (value)
		local tab = #template
		local t = template[tab]
		if #t >= #self.talents[tab].infos then
			t = next()
		end
		t[#t + 1] = value
	end
	
	template[1] = template[1] or {}
		
	for i = 2, code:len() do
		local char = code:sub(i, i)
		if char == stop then
			next()
		else
			index = nmap:find(char, nil, true) - 1
			if not index then return end
			local b = math.fmod(index, 6)
			local a = (index - b) / 6
			
			add(a)
			add(b)
		end
	end

	assert(#template <= #self.talents, "Too many branches")
	do
		for tab, talents in ipairs(self.talents) do
			local t = template[tab] or {}
			template[tab] = t
			for index in ipairs(talents.infos) do
				t[index] = t[index] or 0
			end
		end
	end

	return  template, class
end

local function rtrim(s, c)
	local l = #s
	while l >= 1 and s:sub(l, l) == c do
		l = l - 1
	end
	return s:sub(1, l)
end

function Talented:TemplateToString(template, class, nmap)
	nmap = nmap or talented_map
	
	if not class then
		_, class = UnitClass("player")
	end
	
	local code, ccode = ""
	do
		for index, c in ipairs(classmap) do
			if c == class then
				local i = (index - 1) * 3 + 1
				ccode = nmap:sub(i, i)
				break
			end
		end
	end
	assert(ccode, "invalid class")
	local s = nmap:sub(1, 1)
	for tab, talents in ipairs(self.talents) do
		local tmpl = template[tab]
		for index = 1, #tmpl, 2 do
			local v = tmpl[index] * 6 + (index < #tmpl and tmpl[index + 1] or 0) + 1
			local c = nmap:sub(v, v)
			assert(c)
			code = code .. c
		end
		code = rtrim(code, s)..stop
	end
	return ccode..rtrim(code, "Z")
end

function Talented:PackTemplate(template)
	if not template or template == self.current or template.code then return end
	local res, value = pcall(self.TemplateToString, self, template)
	if res then
		for tab in ipairs(self.talents) do
			template[tab] = nil
		end
		template.code = value
		return true
	end
end

function Talented:UnpackTemplate(template)
	if template == self.current or not template.code then return end
	local res, value = pcall(self.StringToTemplate, self, template.code, nil, nil, template)
	if res then
		template.code = nil
		return true
	end
end
