local LAYOUT_BASE_X = 4
local LAYOUT_BASE_Y = 24
local LAYOUT_OFFSET_X = 64
local LAYOUT_OFFSET_Y = LAYOUT_OFFSET_X
local LAYOUT_DELTA_X = LAYOUT_OFFSET_X / 2
local LAYOUT_DELTA_Y = LAYOUT_OFFSET_Y / 2
local LAYOUT_MAX_COLUMNS = 4
local LAYOUT_MAX_ROWS = 9

local LAYOUT_SIZE_X = LAYOUT_MAX_COLUMNS * LAYOUT_OFFSET_X + LAYOUT_DELTA_X
local LAYOUT_SIZE_Y = LAYOUT_MAX_ROWS * LAYOUT_OFFSET_Y + LAYOUT_DELTA_Y

local LAYOUT_TEXTURE_WIDTH = LAYOUT_SIZE_X / (256+44)
local LAYOUT_TEXTURE_HEIGHT = LAYOUT_SIZE_Y / (256+75)

local dewdrop = AceLibrary("Dewdrop-2.0")
local L = AceLibrary("AceLocale-2.2"):new("Talented")

local COORDS = {
	branch = {
		top = { left = 0.12890625, width = 0.125, height = 0.96875 },
		left = { left = 0.2578125, width = 0.125 },
		-- bottomright = { left = 0.38671875, width = 0.125 },
		-- bottomleft = { left = 0.51171875, width = -0.125 },
		topright = { left = 0.515625, width = 0.125 },
		topleft = { left = 0.640625, width = -0.125 },
		-- tdown = { left = 0.64453125, width = 0.125 },
		-- tup = { left = 0.7734375, width = 0.125 },
	},
	arrow = { 
		top = { left = 0, width = 0.5 },
		left = { left = 0.5, width = 0.5 },
		right = { left = 1.0, width = -0.5 },
	},
}

local GetNumTalentTabs = GetNumTalentTabs
local GetTalentTabInfo = GetTalentTabInfo
local GetNumTalents = GetNumTalents
local GetTalentInfo = GetTalentInfo
local GetTalentPrereqs = GetTalentPrereqs
local min = math.min
local max = math.max
local insert = table.insert


local function SetTextureCoords(object, type, subtype)
	local coords = COORDS[type] and COORDS[type][subtype]
	if not coords then return end
	
	local left = coords.left
	local right = left + coords.width
	local bottom = coords.height or 1
	
	object:SetTexCoord(left, right, 0, bottom)
end

local function MakeButton(parent)
	local button = CreateFrame("Button", nil, parent)
	button:SetTextFontObject(GameFontNormal)
	button:SetHighlightFontObject(GameFontHighlight)
	button:SetDisabledFontObject(GameFontDisable)
	
	local CreateTexture = function(path, blend)
		local t = button:CreateTexture()
		if path then t:SetTexture(path) end
		if blend then t:SetBlendMode(blend) end
		t:SetTexCoord(0, 0.625, 0, 0.6875)
		t:SetAllPoints(button)
		return t
	end
	
	button:SetNormalTexture(CreateTexture("Interface\\Buttons\\UI-Panel-Button-Up"))
	button:SetPushedTexture(CreateTexture("Interface\\Buttons\\UI-Panel-Button-Down"))
	button:SetDisabledTexture(CreateTexture("Interface\\Buttons\\UI-Panel-Button-Disabled"))
	button:SetHighlightTexture(CreateTexture("Interface\\Buttons\\UI-Panel-Button-Highlight", "ADD"))
	
	return button
end

function Talented:CreateBaseButtons(parent)
	local b = MakeButton(parent)
	parent.boptions = b

	local point = function ()
		return "TOPLEFT", "BOTTOMLEFT"
	end
	local onclick = function (this)
		if dewdrop:IsOpen(this) then
			dewdrop:Close()
		else
			dewdrop:Open(this)
		end
	end
	
	b:SetText(L["Options"])
	local width = math.max(110, b:GetTextWidth() + 22)
	b:SetWidth(width)
	b:SetHeight(22)
	b:SetScript("OnClick", function (this)
			dewdrop:Register(this, 
				"children", self.options.args.options,
				"dontHook", true,
				"point", point)
			this:SetScript("OnClick", onclick)
			dewdrop:Open(this)
		end)
	b:SetPoint("TOPLEFT", 14, -4)
	
	b = MakeButton(parent)
	parent.bmode = b

	b:SetText(L["Mode"])
	local width = math.max(110, b:GetTextWidth() + 22)
	b:SetWidth(width)
	b:SetHeight(22)
	b:SetScript("OnClick", function (this)
			dewdrop:Register(this, 
				"children", self.options.args.mode,
				"dontHook", true,
				"point", point)
			this:SetScript("OnClick", onclick)
			dewdrop:Open(this)
		end)
	b:SetPoint("LEFT", parent.boptions, "RIGHT", 14, 0)
	
	b = MakeButton(parent)
	parent.btemplate = b
	
	b:SetText(L["Template"])
	width = math.max(110, b:GetTextWidth() + 22)
	b:SetWidth(width)
	b:SetHeight(22)
	b:SetScript("OnClick", function (this)
			dewdrop:Register(this, 
				"children", self.options.args.template,
				"dontHook", true,
				"point", point)
			this:SetScript("OnClick", onclick)
			dewdrop:Open(this)
		end)
	b:SetPoint("LEFT", parent.bmode, "RIGHT", 14, 0)

	local e = CreateFrame("EditBox", nil, parent)
	parent.editname = e
	e:SetFontObject(ChatFontNormal)
	width = 400
	e:SetWidth(width)
	e:SetHeight(13)
	e:SetAutoFocus(false)
	
	e:SetScript("OnEscapePressed", function (this)
			this:ClearFocus()
		end)
	e:SetScript("OnEditFocusLost", function (this)
			this:SetText(self.template.name)
		end)
	e:SetScript("OnEnterPressed",  function (this)
			self:UpdateTemplateName(self.template, this:GetText())
			self:UpdateView()
			this:ClearFocus()
		end)
	e:SetPoint("LEFT", parent.btemplate, "RIGHT", 14, 0)
end

local function CreateTexture(base, layer, path, blend)
	local t = base:CreateTexture(nil, layer)
	if path then t:SetTexture(path) end
	if blend then t:SetBlendMode(blend)	end
	return t
end

function Talented:EnableUI(enable)
	local f = self.base
	if enable then
		f.boptions:Enable()
		f.bmode:Enable()
		f.btemplate:Enable()
		f.close:Enable()
	else
		dewdrop:Close()
		f.boptions:Disable()
		f.bmode:Disable()
		f.btemplate:Disable()
		f.close:Disable()
	end
end

do
	local gt = GameTooltip
	function Talented:MakeTalentFrame(parent, tab)
		local frame = CreateFrame("Frame", nil, parent)
		frame:SetWidth(LAYOUT_SIZE_X)
		frame:SetHeight(LAYOUT_SIZE_Y)
		frame:SetPoint("TOPLEFT")
		
		
		local t = CreateTexture(frame, "BACKGROUND")
		t:SetWidth(LAYOUT_TEXTURE_WIDTH * 256)
		t:SetHeight(LAYOUT_TEXTURE_HEIGHT * 256)
		t:SetPoint("TOPLEFT")
		frame.topleft = t
		
		t = CreateTexture(frame, "BACKGROUND")
		t:SetWidth(LAYOUT_TEXTURE_WIDTH * 64)
		t:SetHeight(LAYOUT_TEXTURE_HEIGHT * 256)
		t:SetPoint("TOPLEFT", frame.topleft, "TOPRIGHT")
		frame.topright = t
		
		t = CreateTexture(frame, "BACKGROUND")
		t:SetWidth(LAYOUT_TEXTURE_WIDTH * 256)
		t:SetHeight(LAYOUT_TEXTURE_HEIGHT * 128)
		t:SetPoint("TOPLEFT", frame.topleft, "BOTTOMLEFT")
		frame.bottomleft = t
		
		t = CreateTexture(frame, "BACKGROUND")
		t:SetWidth(LAYOUT_TEXTURE_WIDTH * 64)
		t:SetHeight(LAYOUT_TEXTURE_HEIGHT * 128)
		t:SetPoint("TOPLEFT", frame.topleft, "BOTTOMRIGHT")
		frame.bottomright = t
		
		local fs = frame:CreateFontString(nil, "BACKGROUND", "GameFontNormal")
		fs:SetPoint("TOP", 0, -4)
		fs:SetJustifyH("CENTER")
		fs:SetWidth(LAYOUT_SIZE_X)
		frame.name = fs
		
	--~ 		t = CreateTexture(frame, "BORDER")
	--~ 		t:SetWidth(32)
	--~ 		t:SetHeight(32)
	--~ 		t:SetPoint("TOPLEFT")
	--~ 		frame.icon = t

		local overlay = CreateFrame("Frame", nil, frame)
		overlay:SetAllPoints(frame)
		
		frame.overlay = overlay

		local clear = CreateFrame("Button", nil, frame)
		frame.clear = clear
		
		local makeTexture = function (path, blend)
			local t = clear:CreateTexture()
			t:SetAllPoints(clear)
			t:SetTexture(path)
			if blend then
				t:SetBlendMode(blend)
			end
			return t
		end

		clear:SetNormalTexture(makeTexture("Interface\\Buttons\\CancelButton-Up"))
		clear:SetPushedTexture(makeTexture("Interface\\Buttons\\CancelButton-Down"))
		clear:SetHighlightTexture(makeTexture("Interface\\Buttons\\CancelButton-Highlight", "ADD"))
		clear:SetScript("OnClick", function () self:ClearTalentTab(tab) end)
		clear:SetScript("OnEnter", function (self)
				gt:SetOwner(self, "ANCHOR_RIGHT")
				gt:SetText(L["Clear this talent branch"])
			end)
		clear:SetScript("OnLeave", function () 
				gt:Hide()
			end)
		
		clear:SetWidth(32)
		clear:SetHeight(32)
		clear:ClearAllPoints()
		clear:SetPoint("TOPRIGHT", -4, -4)
		
		return frame
	end
end

function Talented:MakeButton(parent, tab, id)
	local button = CreateFrame("Button", nil, parent.child or parent)
	-- ItemButtonTemplate (minus Count and Slot)
	button:SetWidth(37)
	button:SetHeight(37)
	local t = CreateTexture(button, "BORDER")
	t:SetWidth(64)
	t:SetHeight(64)
	t:SetAllPoints(button)
	button.texture = t
	t = CreateTexture(button, nil, "Interface\\Buttons\\UI-Quickslot2")
	t:SetWidth(64)
	t:SetHeight(64)
	t:SetPoint("CENTER", 0, -1)
	button:SetNormalTexture(t)
	t = CreateTexture(button, nil, "Interface\\Buttons\\UI-Quickslot-Depress")
	t:SetWidth(36)
	t:SetHeight(36)
	t:SetPoint("CENTER")
	button:SetPushedTexture(t)
	t = CreateTexture(button, nil, "Interface\\Buttons\\ButtonHilight-Square", "ADD")
	t:SetWidth(36)
	t:SetHeight(36)
	t:SetPoint("CENTER")
	button:SetHighlightTexture(t)
	-- TalentButtonTemplate
	texture = CreateTexture(button, "BACKGROUND", "Interface\\Buttons\\UI-EmptySlot-White")
	texture:SetWidth(64)
	texture:SetHeight(64)
	texture:SetPoint("CENTER", 0, -1)
	button.slot = texture
	
	local makeRank = function (anchor)
		local t = CreateTexture(button, "OVERLAY", "Interface\\Addons\\Talented\\Textures\\border")
		t:SetWidth(32)
		t:SetHeight(32)
		t:SetPoint("CENTER", button, anchor)
		local fs = button:CreateFontString(nil, "OVERLAY", "GameFontNormalSmall")
		fs.texture = t
		fs:SetPoint("CENTER", t)
		return fs
	end
	button.rank = makeRank("BOTTOMRIGHT")
	
	button:RegisterForClicks("LeftButtonUp", "RightButtonUp")

	button:RegisterEvent("CHARACTER_POINTS_CHANGED")
	button:SetScript("OnEvent",
		function (self)
			if GameTooltip:IsOwned(self) then
				GameTooltip:SetTalent(tab, id)
			end
		end)
	button:SetScript("OnEnter",
		function (self)
			GameTooltip:SetOwner(self, "ANCHOR_RIGHT")
			GameTooltip:SetTalent(tab, id)
		end)
	button:SetScript("OnLeave",
		function ()
			GameTooltip:Hide()
		end)
	
	button:SetScript("OnClick", 
		function (self, button, down)
			Talented:OnTalentClick(button, tab, id)
		end)
	
	return button
end

function Talented:MakeBranch(parent)
	local t = CreateTexture(parent, "BORDER", "Interface\\Addons\\Talented\\Textures\\branches-normal")
	t:SetWidth(32)
	t:SetHeight(32)
	t:SetVertexColor(NORMAL_FONT_COLOR.r, NORMAL_FONT_COLOR.g, NORMAL_FONT_COLOR.b)
	return t
end

function Talented:MakeArrow(parent)
	local t = CreateTexture(parent.overlay, "OVERLAY", "Interface\\Addons\\Talented\\Textures\\arrows-normal")
	t:SetWidth(32)
	t:SetHeight(32)
	t:SetVertexColor(NORMAL_FONT_COLOR.r, NORMAL_FONT_COLOR.g, NORMAL_FONT_COLOR.b)
	return t
end

local function offset(row, column)
	return (column - 1) * LAYOUT_OFFSET_X + LAYOUT_DELTA_X, - ((row - 1) * LAYOUT_OFFSET_Y + LAYOUT_DELTA_Y)
end

local function drawline_vertical(list, parent, base_row, base_column, row, column)
	if column ~= base_column then return false end
	for i = row+1, base_row-1 do
		local x, y = offset(i, column)
		local branch = Talented:MakeBranch(parent)
		branch:SetPoint("TOPLEFT", x + 2, y + 32)
		insert(list, branch)
		SetTextureCoords(branch, "branch", "top")
		branch = Talented:MakeBranch(parent)
		branch:SetPoint("TOPLEFT", x + 2, y)
		insert(list, branch)
		SetTextureCoords(branch, "branch", "top")
	end
	local x, y = offset(base_row, base_column)
	local branch = Talented:MakeBranch(parent)
	branch:SetPoint("TOPLEFT", x + 2, y + 32)
	insert(list, branch)
	SetTextureCoords(branch, "branch", "top")
	local arrow = Talented:MakeArrow(parent)
	SetTextureCoords(arrow, "arrow", "top")
	arrow:SetPoint("TOPLEFT", x + 2, y+16)
	insert(list, arrow)
	
	return true
end

local function drawline_horizontal(list, parent, base_row, base_column, row, column)
	if row ~= base_row then return false end
	for i = min(base_column, column)+1, max(base_column, column)-1 do
		local x, y = offset(row, i)
		local branch = Talented:MakeBranch(parent)
		branch:SetPoint("TOPLEFT", x - 32, y - 2)
		insert(list, branch)
		SetTextureCoords(branch, "branch", "left")
		branch = Talented:MakeBranch(parent)
		branch:SetPoint("TOPLEFT", x, y - 2)
		insert(list, branch)
		SetTextureCoords(branch, "branch", "left")
	end
	local x, y = offset(base_row, base_column)
	local branch = Talented:MakeBranch(parent)
	insert(list, branch)
	SetTextureCoords(branch, "branch", "left")
	local arrow = Talented:MakeArrow(parent)
	if base_column < column then
		SetTextureCoords(arrow, "arrow", "right")
		arrow:SetPoint("TOPLEFT", x + 20, y - 2)
		branch:SetPoint("TOPLEFT", x + 32, y - 2)
	else
		SetTextureCoords(arrow, "arrow", "left")
		arrow:SetPoint("TOPLEFT", x - 15, y - 2)
		branch:SetPoint("TOPLEFT", x - 32, y - 2)
	end
	insert(list, arrow)
	return true
end

local function drawline_hv(list, parent, base_row, base_column, row, column)
	local min_row, max_row, min_column, max_column
	
	if base_column < column then
		min_column = base_column + 1
		max_column = column - 1
	else
		min_column = column + 1
		max_column = base_column - 1
	end
	
	for i = min_column, max_column do
		local x, y = offset(row, i)
		local branch = Talented:MakeBranch(parent)
		branch:SetPoint("TOPLEFT", x - 32, y - 2)
		insert(list, branch)
		SetTextureCoords(branch, "branch", "left")
		branch = Talented:MakeBranch(parent)
		branch:SetPoint("TOPLEFT", x, y - 2)
		insert(list, branch)
		SetTextureCoords(branch, "branch", "left")
	end
	
	local x, y = offset(row, base_column)
	local branch = Talented:MakeBranch(parent)
	branch:SetPoint("TOPLEFT", x + 2, y - 2)
	insert(list, branch)
	local branch2 = Talented:MakeBranch(parent)
	SetTextureCoords(branch2, "branch", "left")
	insert(list, branch2)
	if base_column < column then
		branch2:SetPoint("TOPLEFT", x + 35, y - 2)
		SetTextureCoords(branch, "branch", "topleft")
	else
		branch2:SetPoint("TOPLEFT", x - 29, y - 2)
		SetTextureCoords(branch, "branch", "topright")
	end
	
	for i = row+1, base_row-1 do
		local x, y = offset(i, base_column)
		local branch = Talented:MakeBranch(parent)
		branch:SetPoint("TOPLEFT", x + 2, y + 32)
		insert(list, branch)
		SetTextureCoords(branch, "branch", "top")
		branch = Talented:MakeBranch(parent)
		branch:SetPoint("TOPLEFT", x + 2, y)
		insert(list, branch)
		SetTextureCoords(branch, "branch", "top")
	end
	
	x, y = offset(base_row, base_column)
	branch = Talented:MakeBranch(parent)
	branch:SetPoint("TOPLEFT", x + 2, y + 32)
	insert(list, branch)
	SetTextureCoords(branch, "branch", "top")
	local arrow = Talented:MakeArrow(parent)
	SetTextureCoords(arrow, "arrow", "top")
	arrow:SetPoint("TOPLEFT", x + 2, y+16)
	insert(list, arrow)
	
	return true
end

local function drawline_vh(list, parent, base_row, base_column, row, column)
	return true
end

local function drawline(list, parent, base_row, base_column, row, column)
	return drawline_vertical(list, parent, base_row, base_column, row, column) or
		drawline_horizontal(list, parent, base_row, base_column, row, column) or
		drawline_hv(list, parent, base_row, base_column, row, column) or
		drawline_vh(list, parent, base_row, base_column, row, column)
end

function Talented:CreateTalentMatrix(parent)
	local talents = self.talents
	local numtabs = GetNumTalentTabs()
	for tab = 1, numtabs do
		local frame = self:MakeTalentFrame(parent, tab)
		local name, texture, _, background = GetTalentTabInfo(tab)
		background = "Interface\\TalentFrame\\"..background.."-"
		frame.topleft:SetTexture(background.."TopLeft")
		frame.topright:SetTexture(background.."TopRight")
		frame.bottomleft:SetTexture(background.."BottomLeft")
		frame.bottomright:SetTexture(background.."BottomRight")
		if frame.icon then frame.icon:SetTexture(texture) end
		local matrix = { infos = {}, frame = frame, name = name }
		talents[tab] = matrix
		local infos = matrix.infos
		local numtalents = GetNumTalents(tab)
		for idx = 1, numtalents do
			local name, texture, row, column, _, ranks, exceptional = GetTalentInfo(tab, idx)
			local button = self:MakeButton(frame, tab, idx)

			infos[idx] = {
				index = idx,
				name = name,
				texture = texture,
				row = row,
				column = column,
				ranks = ranks,
				exceptional = exceptional,
				button = button,
			}
			infos[column * 100 + row] = infos[idx]
			
			button:SetPoint("TOPLEFT", offset(row, column))
			button.rank:SetText(ranks)
			button.texture:SetTexture(texture)
			if exceptional then
				button.slot:SetVertexColor(1, 1, 0)
			else
				button.slot:SetVertexColor(0, 1, 0)
			end
		end
		for idx = 1, numtalents do
			local _, _, row, column = GetTalentInfo(tab, idx)
			local reqs = {}
			local function prereq(...)
				for i=1, select("#", ...), 3 do
					local nrow, ncol = select(i, ...)
					local req = { source = infos[ncol * 100 + nrow], elements = {} }
					drawline(req.elements, frame, row, column, nrow, ncol)
					insert(reqs, req)
				end
			end
			prereq(GetTalentPrereqs(tab, idx))
			if next(reqs) then
				infos[idx].reqs = reqs
			end
		end
		frame:SetPoint("TOPLEFT", (tab-1) * LAYOUT_SIZE_X + LAYOUT_BASE_X, -LAYOUT_BASE_Y)
	end
	parent:SetWidth(numtabs * LAYOUT_SIZE_X + LAYOUT_BASE_X * 2)
	parent:SetHeight(LAYOUT_SIZE_Y + LAYOUT_BASE_Y + LAYOUT_BASE_X)
	parent:SetPoint("CENTER")
end

function Talented:CreateBaseFrame()
	local frame = self.base
	if frame then return frame end
	
	local onshow = function ()
		SetButtonPulse(TalentMicroButton, 0, 1)
		PlaySound("TalentScreenOpen")
		UpdateMicroButtons()
	end
	
	frame = CreateFrame("Frame", "TalentedFrame", UIParent)
	frame:SetFrameStrata("DIALOG")
	frame:EnableMouse(true)
	frame:SetMovable(true)
	frame:SetToplevel(true)
	frame:SetWidth(50)
	frame:SetHeight(50)
	frame:SetBackdrop({
		bgFile = "Interface\\TutorialFrame\\TutorialFrameBackground",
		edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border",
		tile = true,
		edgeSize = 16,
		tileSize = 32,
		insets = {
			left = 5,
			right = 5,
			top = 5,
			bottom = 5
		}
	})

	frame:SetScript("OnMouseDown", function(self) self:StartMoving() end)
	frame:SetScript("OnMouseUp", function(self) self:StopMovingOrSizing() end)

	frame.points = frame:CreateFontString(nil, "OVERLAY", "GameFontNormal")
	frame.points:SetJustifyH("RIGHT")
	frame.points:SetWidth(80)
	frame.points:SetHeight(14)
	frame.points:SetPoint("TOPRIGHT", frame, "TOPRIGHT", -40, -6)
	
	local close = CreateFrame("Button", nil, frame)
	frame.close = close
	
	local makeTexture = function (path, blend)
		local t = close:CreateTexture()
		t:SetAllPoints(close)
		t:SetTexture(path)
		if blend then
			t:SetBlendMode(blend)
		end
		return t
	end

	close:SetNormalTexture(makeTexture("Interface\\Buttons\\UI-Panel-MinimizeButton-Up"))
	close:SetPushedTexture(makeTexture("Interface\\Buttons\\UI-Panel-MinimizeButton-Down"))
	close:SetHighlightTexture(makeTexture("Interface\\Buttons\\UI-Panel-MinimizeButton-Highlight", "ADD"))
	close:SetScript("OnClick", function () HideUIPanel(frame) end)
	
	close:SetWidth(32)
	close:SetHeight(32)
	close:ClearAllPoints()
	close:SetPoint("TOPRIGHT", 1, 0)

	self.base = frame
	self:CreateBaseButtons(frame)
	self:CreateTalentMatrix(frame)
--~ 	frame:SetScale(0.5)

	frame:SetScript("OnShow", onshow)
	frame:SetScript("OnHide", function(self)
			PlaySound("TalentScreenClose")
			if dewdrop:IsOpen(self.boptions) or dewdrop:IsOpen(self.bmode) or dewdrop:IsOpen(self.btemplate) then
				dewdrop:Close()
			end
			if self.mode == "apply" then
				self:Print(L["Error! Talented window has been closed during template application. Please reapply later."])
				self.mode = "view"
				self:EnableUI()
			end
		end)

	UIPanelWindows["TalentedFrame"] = { area = "center", pushable = 0, whileDead = 1 }
	onshow()
	return frame
end

